﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Operators {
  using CloneMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItem>;
  using TraceMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItemList<HeuristicLab.Core.IItem>>;

  /// <summary>
  /// An operator that tracks parents after crossover.
  /// </summary>
  [Item("AfterCrossoverOperator", "An operator that tracks parents after crossover.")]
  [StorableClass]
  public class AfterCrossoverOperator : SingleSuccessorOperator {
    public const string ParentsParameterName = "Parents";
    public const string ChildParameterName = "Child";
    private const string TargetName = "TSPTour";

    #region Parameter properties
    public ILookupParameter<ItemArray<Permutation>> ParentsParameter {
      get { return (ScopeTreeLookupParameter<Permutation>)Parameters[ParentsParameterName]; }
    }
    public ILookupParameter<Permutation> ChildParameter {
      get { return (ILookupParameter<Permutation>)Parameters[ChildParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[CreateTrackingDatastructures.GlobalCloneMapParameterName]; }
    }
    public LookupParameter<TraceMapType> GlobalTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[CreateTrackingDatastructures.GlobalTraceMapParameterName]; }
    }
    #endregion

    #region Properties
    public CloneMapType GlobalCloneMap {
      get { return GlobalCloneMapParameter.ActualValue; }
    }
    public TraceMapType GlobalTraceMap {
      get { return GlobalTraceMapParameter.ActualValue; }
    }
    public ItemArray<Permutation> Parents {
      get { return ParentsParameter.ActualValue; }
    }
    public Permutation Child {
      get { return ChildParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private AfterCrossoverOperator(bool deserializing) : base(deserializing) { }
    private AfterCrossoverOperator(AfterCrossoverOperator original, Cloner cloner) : base(original, cloner) { }
    public AfterCrossoverOperator()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<Permutation>(ParentsParameterName, "The parent permutations which should be crossed."));
      Parameters.Add(new LookupParameter<Permutation>(ChildParameterName, "The child permutation resulting from the crossover."));
      Parameters.Add(new LookupParameter<CloneMapType>(CreateTrackingDatastructures.GlobalCloneMapParameterName, "A global map keeping track of trees and their clones (made during selection)."));
      Parameters.Add(new LookupParameter<TraceMapType>(CreateTrackingDatastructures.GlobalTraceMapParameterName, "A global cache containing tracing info."));
      ParentsParameter.ActualName = TargetName;
      ChildParameter.ActualName = TargetName;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AfterCrossoverOperator(this, cloner);
    }

    public override IOperation Apply() {
      var originalParents = new ItemList<IItem>(ParentsParameter.ActualValue.Select(x => GlobalCloneMap[x]));
      GlobalTraceMap[Child] = originalParents;

      return base.Apply();
    }
  }
}
