﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Algorithms.ScatterSearch.TravelingSalesman;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Analysis.AlgorithmBehavior {

  /// <summary>
  /// An operator that analyzes the quality of schemas.
  /// </summary>
  [Item("ConvergenceSpeedAnalyzer", "An operator that calculates the convergence speed.")]
  [StorableClass]
  public sealed class ConvergenceSpeedAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphResultParameterName = "PopulationGraph";
    private const string SchemataParameterName = "Schemata";
    private const string SchemaOccurenceInGenerationsMatrixParameterName = "ConvergenceSpeedAnalyzer";
    private const string QualitiesParameterName = "Qualities";
    private const string DistanceMatrixParameterName = "DistanceMatrix";
    private const string GenerationsParameterName = "Generations";
    private const string BestSolutionParameterName = "Best TSP Solution";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<DistanceMatrix> DistanceMatrixParameter {
      get { return (ILookupParameter<DistanceMatrix>)Parameters[DistanceMatrixParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public DistanceMatrix DistanceMatrix {
      get { return DistanceMatrixParameter.ActualValue; }
    }
    public HeuristicLab.Analysis.DataTable Qualities {
      get { return ((HeuristicLab.Analysis.DataTable)ResultsParameter.ActualValue[QualitiesParameterName].Value); }
    }
    public PathTSPTour BestSolution {
      get { return ((PathTSPTour)ResultsParameter.ActualValue[BestSolutionParameterName].Value); }
    }
    #endregion

    [StorableConstructor]
    private ConvergenceSpeedAnalyzer(bool deserializing) : base(deserializing) { }
    private ConvergenceSpeedAnalyzer(ConvergenceSpeedAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public ConvergenceSpeedAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<DistanceMatrix>(DistanceMatrixParameterName, "The distance matrix of the TSP."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ConvergenceSpeedAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var graph = (GenealogyGraph<Permutation>)Results[PopulationGraphResultParameterName].Value;
      Dictionary<int, double> occurences = new Dictionary<int, double>();

      for (int i = 0; i < GenerationsParameter.ActualValue.Value; i++) {
        var individuals = graph.Values.Where(x => x.Rank.Contains(i) || x.Rank.Contains(i + 0.5));

        double convInGeneration = 0.0;
        foreach (var ind1 in individuals) {
          double convForInd = 0.0;
          foreach (var ind2 in individuals) {
            convForInd += TravelingSalesmanSimilarityCalculator.CalculateDiversity((Permutation)ind1.Data, (Permutation)ind2.Data);
          }
          convForInd = convForInd / individuals.Count();
          convInGeneration += convForInd;
        }
        convInGeneration = convInGeneration / individuals.Count();
        occurences.Add(i, convInGeneration);
      }

      var schemataMatrix = new Dictionary<int, string>();
      foreach (var ind in occurences.Keys) {
        if (!schemataMatrix.ContainsKey(ind)) {
          string result = occurences[ind].ToString();
          schemataMatrix.Add(ind, result);
        }
      }

      var stringMatrix = new StringMatrix(schemataMatrix.Keys.Count, 2);
      for (int i = 0; i < schemataMatrix.Keys.Count; i++) {
        var element = schemataMatrix.ElementAt(i);
        stringMatrix[i, 0] = element.Key.ToString();
        stringMatrix[i, 1] = element.Value;
      }
      stringMatrix.ColumnNames = new[] { "Generation", "Diversity" };

      Results.Add(new Result(SchemaOccurenceInGenerationsMatrixParameterName, stringMatrix));

      return base.Apply();
    }
  }
}

