﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Analysis.AlgorithmBehavior {

  /// <summary>
  /// An operator that analyzes the quality of schemas.
  /// </summary>
  [Item("SchemaOccurenceInGenerationsAnalyzer", "An operator that analyzes broken inheritance of building blocks.")]
  [StorableClass]
  public sealed class SchemaOccurenceInGenerationsAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphResultParameterName = "PopulationGraph";
    private const string SchemataParameterName = "Schemata";
    private const string SchemaOccurenceInGenerationsMatrixParameterName = "SchemaOccurenceInGenerationsMatrix";
    private const string QualitiesParameterName = "Qualities";
    private const string DistanceMatrixParameterName = "DistanceMatrix";
    private const string GenerationsParameterName = "Generations";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<DistanceMatrix> DistanceMatrixParameter {
      get { return (ILookupParameter<DistanceMatrix>)Parameters[DistanceMatrixParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public DistanceMatrix DistanceMatrix {
      get { return DistanceMatrixParameter.ActualValue; }
    }
    public HeuristicLab.Analysis.DataTable Qualities {
      get { return ((HeuristicLab.Analysis.DataTable)ResultsParameter.ActualValue[QualitiesParameterName].Value); }
    }
    #endregion

    [StorableConstructor]
    private SchemaOccurenceInGenerationsAnalyzer(bool deserializing) : base(deserializing) { }
    private SchemaOccurenceInGenerationsAnalyzer(SchemaOccurenceInGenerationsAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public SchemaOccurenceInGenerationsAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<DistanceMatrix>(DistanceMatrixParameterName, "The distance matrix of the TSP."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SchemaOccurenceInGenerationsAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var graph = (GenealogyGraph<Permutation>)Results[PopulationGraphResultParameterName].Value;
      var schemata = (ItemArray<IntArray>)Results[SchemataParameterName].Value;

      var schemataMatrix = new Dictionary<IntArray, string>(new SchemaEqualityComparer());
      foreach (var schema in schemata) {
        if (!schemataMatrix.ContainsKey(schema)) {
          string result = ExploreSchemaCountOverGenerations(graph, schema);
          schemataMatrix.Add(schema, result);
        }
      }

      // order schemas according to their global occurrence count
      var sortedSchemataMatrix = schemataMatrix.OrderByDescending(x => x.Value[0]).ToDictionary(x => x.Key, x => x.Value);

      var stringMatrix = new StringMatrix(sortedSchemataMatrix.Keys.Count, 2);
      for (int i = 0; i < sortedSchemataMatrix.Keys.Count; i++) {
        var element = sortedSchemataMatrix.ElementAt(i);
        stringMatrix[i, 0] = new Permutation(PermutationTypes.RelativeUndirected, element.Key).ToString();
        stringMatrix[i, 1] = element.Value;
      }
      stringMatrix.ColumnNames = new[] { "Schema", "SchemaCountOverGeneration" };

      Results.Add(new Result(SchemaOccurenceInGenerationsMatrixParameterName, stringMatrix));

      return base.Apply();
    }

    private string ExploreSchemaCountOverGenerations(GenealogyGraph<Permutation> graph, IntArray schema) {
      int[] occurences = new int[GenerationsParameter.ActualValue.Value + 1];

      foreach (var individual in graph.Values) {
        if (AlgorithmBehaviorHelpers.IsSubtour(schema, (Permutation)individual.Data)) {
          foreach (var rank in individual.Rank) {
            occurences[(int)rank]++;
          }
        }
      }

      StringBuilder builder = new StringBuilder();
      foreach (int x in occurences)
        builder.Append(x.ToString() + ", ");

      return builder.ToString();
    }
  }
}
